@@ -38,7 +38,7 @@
     };
 
     function wrapAnchor(anchor, model, icon) {
-        return $("<li />", {"data-icon": icon}).append(anchor.data("model", model));
+        return $("<li />", {"data-icon": icon}).append(anchor.mopidyModel(model));
     }
 
     function getAlbumImageURI(album) {
@@ -124,12 +124,12 @@
 
         _listitem: function(model) {
             var $li = $.mopidy.itemCallbacks[model.__model__](model);
-            $li.data("model", model);
+            $li.mopidyModel(model);
             return $li;
         },
 
         mapKey: function(array, key) {
-            return $.map(array, function(e) { return e[key]; });
+            return $.map(array || [], function(e) { return e[key]; });
         },
 
         swipe: function($element, direction, callback) {
@@ -192,197 +192,363 @@
     });
 })( jQuery, this );
 
+(function($, undefined) {
+
+    $.extend($.mopidy, {
+
+        isRef: function(model, type) {
+            return model.__model__ === "Ref" && (type === undefined || model.type === type);
+        },
+
+        isAlbum: function(model) {
+            return model.__model__ === "Album";
+        },
+
+        isArtist: function(model) {
+            return model.__model__ === "Artist";
+        },
+
+        isTrack: function(model) {
+            return model.__model__ === "Track";
+        },
+
+        isPlaylist: function(model) {
+            return model.__model__ === "Playlist";
+        },
+
+        getModelByUri: function(models, uri) {
+            for (var i = 0; i !== models.length; ++i) {
+                if ((models[i].track || models[i]).uri === uri) {
+                    return models[i];
+                }
+            }
+            return null;
+        },
+
+    });
+
+    $.extend($.fn, {
+
+        mopidyModel: function(value) {
+	    if (value === undefined) {
+		return this.data("model");
+	    } else {
+		return this.data("model", value);
+	    }
+	},
+
+	mopidyRemoveModel: function() {
+	    return this.removeData("model");
+	},
+
+    });
+
+})(jQuery);
+
 (function( $, window, undefined ) {
 
+    function defaultRefClickCallback() {
+        var $item = $(this),
+            $library = $item.closest(":jqmData(role='library')"),
+            parentUri = $library.library("getCurrentUri"),
+            uri = $item.mopidyModel().uri;
+        $.mopidy.connect().then(function() {
+            var tracklist = this.tracklist,
+                playback = this.playback;
+            tracklist.clear().then(function() {
+                tracklist.add(null, null, parentUri).then(function(tlTracks) {
+                    var tlTrack = $.mopidy.getModelByUri(tlTracks, uri);
+                    if (tlTrack) {
+                        return playback.play(tlTrack);
+                    }
+                    tracklist.add(null, null, uri).then(function(tlTracks) {
+                        return playback.play(tlTracks[0]);
+                    });
+                    $item.nextAll().each(function() {
+                        var model = $(this).mopidyModel();
+                        if ($.mopidy.isRef(model, "track")) {
+                            tracklist.add(null, null, model.uri);
+                        }
+                    });
+                    $item.prevAll().each(function() {
+                        var model = $(this).mopidyModel();
+                        if ($.mopidy.isRef(model, "track")) {
+                            tracklist.add(null, 0, model.uri);
+                        }
+                    });
+                });
+            });
+        });
+    }
+
+    function defaultAlbumClickCallback() {
+        var $item = $(this),
+            album = $item.mopidyModel();
+        $.mopidy.connect().then(function() {
+            var tracklist = this.tracklist,
+                playback = this.playback;
+            tracklist.clear().then(function() {
+                tracklist.add(null, null, album.uri).then(function(tlTracks) {
+                    playback.play(tlTracks[0]);
+                });
+            });
+        });
+    }
+
+    function defaultArtistClickCallback() {
+        var $item = $(this),
+            artist = $item.mopidyModel();
+        $.mopidy.connect().then(function() {
+            var tracklist = this.tracklist,
+                playback = this.playback;
+            tracklist.clear().then(function() {
+                tracklist.add(null, null, artist.uri).then(function(tlTracks) {
+                    playback.play(tlTracks[0]);
+                });
+            });
+        });
+    }
+
+    function defaultTrackClickCallback() {
+        var $item = $(this),
+            track = $item.mopidyModel(),
+            prevTracks = $item.prevAll().map(function() {
+                var model = $(this).mopidyModel();
+                return $.mopidy.isTrack(model) ? model : null;
+            }).get(),
+            nextTracks = $item.nextAll().map(function() {
+                var model = $(this).mopidyModel();
+                return $.mopidy.isTrack(model) ? model : null;
+            }).get(),
+            tracks = [].concat(prevTracks, track, nextTracks);
+        $.mopidy.connect().then(function() {
+            var tracklist = this.tracklist,
+                playback = this.playback;
+            tracklist.clear().then(function() {
+                tracklist.add(tracks).then(function(tlTracks) {
+                    playback.play($.mopidy.getModelByUri(tlTracks, track.uri));
+                });
+            });
+        });
+    }
+
+    function defaultRefItemCallback(ref) {
+        var icon = ref.type === "track" ? false : undefined;
+        return $("<li />", {"data-icon": icon})
+            .append($("<a />", {href: ref.uri, text: ref.name || ""}))
+            .mopidyModel(ref);
+    }
+
+    function defaultAlbumItemCallback(album) {
+        var artists = $.mopidy.mapKey(album.artists, "name").join(", ");
+        return $("<li />", {"data-icon": false})
+            .append(
+                $("<a />", {href: album.uri})
+                    .append($("<h1 />", {text: album.name || ""}))
+                    .append($("<p />", {text: artists}))
+            )
+            .mopidyModel(album);
+    }
+
+    function defaultArtistItemCallback(artist) {
+        return $("<li />", {"data-icon": false})
+            .append($("<a />", {href: artist.uri, text: artist.name || ""}))
+            .mopidyModel(artist);
+    }
+
+    function defaultTrackItemCallback(track) {
+        var album = track.album || {},
+            artists = $.mopidy.mapKey(album.artists, "name").join(", "),
+            text = [];
+        if (artists.length) {
+            text.push(artists);
+        }
+        if (album.name) {
+            text.push(album.name);
+        }
+        return $("<li />", {"data-icon": false})
+            .append(
+                $("<a />", {href: track.uri})
+                    .append($("<h1 />", {text: track.name || ""}))
+                    .append($("<p />", {text: text.join(" - ")}))
+            )
+            .mopidyModel(track);
+    }
+
     $.widget( "mopidy.library", $.mobile.listview, {
+
 	initSelector: ":jqmData(role='library')",
 
         options: {
-            model: null,
-            history: [],
-            uri: null
-        },
-
-        back: function() {
-            var model = this.options.history.pop();
-            if ( model !== undefined ) {
-                this._browse( model );
-            }
+            back: null,
+            forward: null,
+            title: null  // label, caption?
         },
 
         _create: function() {
+            var $element = this.element,
+                options = this.options,
+                widget = this;
             this._super();
-            var library = this,
-                $library = this.element,
-                add = $library.data( "add" ),
-                back = $library.data( "back" ),
-                title = $library.data( "title" );
+            this._history = [];
+            this._currentHistoryIndex = -1;
 
-            if ( add ) {
-                $( add ).click(function() {
-                    var uris = [];
-                    $library.find( "li" ).each(function() {
-                        var $this = $( this ),
-                            model = $this.data( "model" );
-                        if (model.type !== "directory") {
-                            uris.push(model.uri);
+            this._on({
+                "click li a": function(event) {
+                    var $item = $(event.currentTarget).closest("li"),
+                        model = $item.mopidyModel();
+                    if ($.mopidy.isAlbum(model)) {
+                        defaultAlbumClickCallback.call($item.get(), model);
+                    } else if ($.mopidy.isArtist(model)) {
+                        defaultArtistClickCallback.call($item.get(), model);
+                    } else if ($.mopidy.isTrack(model)) {
+                        defaultTrackClickCallback.call($item.get(), model);
+                    } else if ($.mopidy.isRef(model, "track")) {
+                        defaultRefClickCallback.call($item.get(), model);
+                    } else {
+                        $.mopidy.connect().then(function() {
+                            widget._browse(this, model.uri);
+                            widget._pushState(null, model.name, model.uri);
+                        });
+                    }
+                    event.preventDefault();
+                },
+                "filterablebeforefilter": function(event, data) {
+                    var $input = $(data.input),
+                        query = $input.val(),
+                        uri = widget.getCurrentUri();
+                    $.mopidy.connect().then(function() {
+                        if (query && query.length > 0) {
+                            widget._search(this, {any: [query]}, uri ? [uri] : null);
+                        } else {
+                            widget._browse(this, uri);
                         }
                     });
-                    if (uris.length) {
-                        $.mopidy.connect().then(function() {
-                            var mopidy = this,
-                                i;
-                            mopidy.tracklist.add(null, null, uris[0]).then(function(tlTracks) {
-                                if (tlTracks.length) {
-                                    mopidy.playback.play(tlTracks[0]);
-                                }
-                            });
-                            for (i = 1; i !== uris.length; ++i) {
-                                mopidy.tracklist.add(null, null, uris[i]);
-                            }
-                        });
-                    }
-                });
-            }
-
-            if ( back ) {
-                $( back ).click(function() {
-                    $library.parent().find( ".ui-input-search input" ).val( "" );
-                    library.back();
-                });
-            }
-            if ( title ) {
-                library.$title = $(title);
-                library.deftitle = $(title).text();
-            }
-            $library.on( "filterablebeforefilter", function ( e, data ) {
-                e.preventDefault();
-                var $input = $( data.input ),
-                    query = $input.val();
-                if ( query && query.length > 0 ) {
-                    library._search( query );
-                } else {
-                    library._browse( library.options.model );
+                    event.preventDefault();
+                    event.stopPropagation();
                 }
             });
-            $library.on( "click", "a", function( event ) {
-                var $a = $(this),
-                    $li = $a.parent();
-                event.preventDefault();
-                $.mopidy.connect().then(function() {
-                    var mopidy = this,
-                        model = $li.data("model");
-                    if (model.type === "directory") {
-                        library.options.history.push( library.options.model );
-                        library._browse( model );
-                    } else if (model.uri) {
-                        mopidy.tracklist.add(null, null, model.uri).then(function(tlTracks) {
-                            mopidy.playback.play(tlTracks[0]);
-                        });
-                    }
+
+            // FIXME: _setOption, _init?
+            $.each(["back", "forward"], function(i, option) {
+                $(options[option]).on("click", function() {
+                    $element.library(option);
                 });
             });
 
-            $.mopidy.connect().then(function() {
-                console.log("library:online");
-                library._browse( null );
-            });
-	},
+            this._pushState(null, $(options.title).text(), null);
+            this.go(0);
+        },
 
-        _browse: function( model ) {
-            var library = this,
-                $library = this.element,
-                uri = model ? model.uri : null;
-            $.mopidy.connect().then(function() {
-                this.library.browse( uri ).then(function( refs ) {
-                    var i;
-                    if ( library.$title ) {
-                        library.$title.text( model ? model.name : library.deftitle );
-                    }
-                    library.options.model = model;
-                    library.options.uri = uri;
-                    $library.empty();
-                    for ( i = 0; i !== refs.length; ++i ) {
-                        $library.append( $.mopidy._listitem( refs[i] ) );
-                    }
-                    $library.library( "refresh" );
-                    $library.trigger( "updatelayout");
-                });
+        _pushState: function(state, title, uri) {
+            var options = this.options,
+                index = this._currentHistoryIndex + 1;
+            $(options.back).toggle(index !== 0);
+            $(options.forward).hide();
+            $(options.title).text(title || "");
+            this._history.splice(index, Number.MAX_VALUE, {
+                state: state,
+                title: title,
+                uri: uri
             });
+            this._currentHistoryIndex = index;
+        },
+
+        _browse: function(mopidy, uri) {
+            var $element = this.element;
+            mopidy.library.browse(uri).then(function(refs) {
+                var $items = $.map(refs, defaultRefItemCallback);
+                $element.empty().append($items).library("refresh");
+            });
+        },
+
+        _search: function(mopidy, query, uris, exact) {
+            var $element = this.element,
+                promise;
+            if (exact) {
+                promise = mopidy.library.findExact(query, uris);
+            } else {
+                promise = mopidy.library.search(query, uris);
+            }
+            promise.then(function(results) {
+                var $items = $.map(results || [], function(result) {
+                    return [].concat(
+                        $.map(result.artists || [], defaultArtistItemCallback),
+                        $.map(result.albums || [], defaultAlbumItemCallback),
+                        $.map(result.tracks || [], defaultTrackItemCallback)
+                    );
+                });
+                $element.empty().append($items).library("refresh");
+            });
+        },
+
+        back: function() {
+            this.go(-1);
         },
 
         browse: function(uri) {
-            uri;
+            var widget = this;
+            $.mopidy.connect().then(function() {
+                widget._browse(this, uri);
+                widget._pushState(null, null, uri);
+            });
         },
 
         findExact: function(query, uris) {
-            query;
-            uris;
-        },
-
-        _lookup: function(uri) {
-            uri;
-        },
-
-        //refresh: function(uri) {
-        //    if (uri !== undefined) {
-        //        $.mopidy.connect().then(function() {
-        //            this.library.refresh(uri);
-        //        });
-        //    }
-        //    this._super();
-        //},
-
-        search: function(query, uris) {
-            query;
-            uris;
-        },
-
-        _search: function( query ) {
-            var $library = this.element,
-                uris = this.options.uri ? [this.options.uri] : undefined;
+            var widget = this;
             $.mopidy.connect().then(function() {
-                this.library.search( { any: [ query ] }, uris ).then(function( results ) {
-                    $library.empty();
-                    $.each( results, function() {
-                        var i;
-                        if ( this.artists ) {
-                            for ( i = 0; i !== this.artists.length; ++i ) {
-                                $library.append( $.mopidy._listitem( this.artists[i] ) );
-                            }
-                        }
-                        if ( this.albums ) {
-                            for (i = 0; i !== this.albums.length; ++i) {
-                                $library.append( $.mopidy._listitem( this.albums[i]) );
-                            }
-                        }
-                        if ( this.tracks ) {
-                            for (i = 0; i !== this.tracks.length; ++i) {
-                                $library.append( $.mopidy._listitem( this.tracks[i]) );
-                            }
-                        }
-                    });
-
-                    $library.library( "refresh" );
-                    $library.trigger( "updatelayout");
-                });
+                widget._search(this, query, uris, true);
+                //widget._pushState({query: query, uris: uris, exact: true});
             });
         },
+
+        forward: function() {
+            this.go(+1);
+        },
+
+        getCurrentUri: function() {
+            return this._history[this._currentHistoryIndex].uri;
+        },
+
+
+        go: function(delta) {
+            var options = this.options,
+                history = this._history,
+                index = this._currentHistoryIndex + delta,
+                widget = this;
+            if (index >= 0 && index < history.length) {
+                $.mopidy.connect().then(function() {
+                    widget._browse(this, history[index].uri);
+                });
+                $(options.back).toggle(index > 0);
+                $(options.forward).toggle(index < history.length - 1);
+                $(options.title).text(history[index].title || "");
+                this._currentHistoryIndex = index;
+            }
+        },
+
+        refresh: function(/*uri*/) {
+            // if (uri !== undefined) {
+            //     $.mopidy.connect().then(function() {
+            //         this.library.refresh(uri);
+            //     });
+            // }
+            this._super();
+        },
+
+        search: function(query, uris) {
+            var widget = this;
+            $.mopidy.connect().then(function() {
+                widget._search(this, query, uris, false);
+                //widget._pushState({query: query, uris: uris, exact: false});
+            });
+        },
+
     });
-})( jQuery, this );
+})(jQuery, this);
 
 (function ($, window, undefined) {
 
-    // TODO:
-    // - define & trigger change events
-    // - toggle control (play/pause)
-    // - update volume/seek on events (but not if source)
-    // - handle non-slider input for volume/seek
-    // - hide/disable controls if not available for current state
-    // - refresh/trigger "updatelayout" on hide/show?
-    // - refresh (check for new controls)?
-    // - calculate/update timePosition?
-
     $.widget("mopidy.playback", {
 
         initSelector: ":jqmData(role='playback')",
@@ -499,8 +665,6 @@
                         $this.prop("disabled", state === "paused");
                         break;
                     case "play":
-                        // TODO: check state chart!
-                        $this.prop("disabled", state === "playing");
                         break;
                     case "resume":
                         $this.prop("disabled", state !== "paused");
@@ -650,8 +814,27 @@
 
 (function($, window, undefined) {
 
-    function defaultAnchorCallback(track) {
-        return $("<a />", {href: track.uri, text: track.name || ""});
+    function defaultTrackClickCallback() {
+        var $item = $(this),
+            $playlist = $item.closest(".ui-collapsible"),
+            tracks = $playlist.mopidyModel().tracks,
+            uri = $item.mopidyModel().uri;
+        $.mopidy.connect().then(function() {
+            var tracklist = this.tracklist,
+                playback = this.playback;
+            tracklist.clear().then(function() {
+                tracklist.add(tracks).then(function(tlTracks) {
+                    var tlTrack = $.mopidy.getModelByUri(tlTracks, uri);
+                    playback.play(tlTrack || tlTracks[0]);
+                });
+            });
+        });
+    }
+
+    function defaultTrackItemCallback(track) {
+        return $("<li />", {"data-icon": false})
+            .append($("<a />", {href: track.uri, text: track.name || ""}))
+            .mopidyModel(track);
     }
 
     $.widget("mopidy.playlists", $.mobile.collapsibleset, {
@@ -659,7 +842,6 @@
 	initSelector: ":jqmData(role='playlists')",
 
         options: {
-            anchorCallback: defaultAnchorCallback
         },
 
         _create: function() {
@@ -668,19 +850,8 @@
 
             this._on({
                 "click li a": function(event) {
-                    var $item = $(event.currentTarget).closest("li"),
-                        $playlist = $item.closest(".ui-collapsible"),
-                        tracks = $playlist.data("model").tracks,
-                        index = $item.index();
-                    $.mopidy.connect().then(function() {
-                        var tracklist = this.tracklist,
-                            playback = this.playback;
-                        tracklist.clear().then(function() {
-                            tracklist.add(tracks).then(function(tlTracks) {
-                                playback.play(tlTracks[index]);
-                            });
-                        });
-                    });
+                    var $item = $(event.currentTarget).closest("li");
+                    defaultTrackClickCallback.call($item.get(), event);
                     event.preventDefault();
                 }
             });
@@ -703,19 +874,14 @@
                 options = this.options;
             mopidy.playlists.getPlaylists().then(function(playlists) {
                 var $collapsibles = $.map(playlists, function(playlist) {
-                    var $items = $.map(playlist.tracks || [], function(track) {
-                        return $("<li />", {"data-icon": false})
-                            .append(options.anchorCallback(track))
-                            .data("model", track);
-                    });
-                    return $("<div />")
+                    var $items = $.map(playlist.tracks || [], defaultTrackItemCallback);
+                    return $("<div />").jqmData("role", "collapsible")
                         .append($("<h1 />", {text: playlist.name || ""}))
                         .append($("<ul />").append($items).listview())
                         .collapsible({inset: options.inset})
-                        .data("model", playlist);
+                        .mopidyModel(playlist);
                 });
-                $element.empty().append($collapsibles);
-                $element.playlists("refresh");
+                $element.empty().append($collapsibles).playlists("refresh");
             });
         },
 
@@ -731,12 +897,12 @@
             });
         },
 
-        refresh: function(uriScheme) {
-            if (uriScheme !== undefined) {
-                $.mopidy.connect().then(function() {
-                    this.playlists.refresh(uriScheme);
-                });
-            }
+        refresh: function(/*uriScheme*/) {
+            //if (uriScheme !== undefined) {
+            //    $.mopidy.connect().then(function() {
+            //        this.playlists.refresh(uriScheme);
+            //    });
+            //}
             this._super();
         },
 
@@ -798,7 +964,7 @@
             this._on({
                 "click li a": function(event) {
                     var $item = $(event.currentTarget).closest("li"),
-                        model = $item.data("model");
+                        model = $item.mopidyModel();
                     $.mopidy.connect().then(function() {
                         this.playback.play(model);
                     });
@@ -806,7 +972,7 @@
                 },
                 "swipeleft li": function(event) {
                     var $item = $(event.currentTarget),
-                        model = $item.data("model");
+                        model = $item.mopidyModel();
                     $.mopidy.swipe($item, "left", function() {
                         widget.remove({tlid: [model.tlid]});
                     });
@@ -847,7 +1013,7 @@
                 var $items = $.map(tlTracks, function(tlTrack) {
                     return $("<li />", {"data-icon": false})
                         .append(anchorCallback(tlTrack))
-                        .data("model", tlTrack);
+                        .mopidyModel(tlTrack);
                 });
                 $element.empty().append($items).tracklist("refresh");
                 widget._updateCurrentTlId(widget._currentTlId);
@@ -874,7 +1040,7 @@
         _updateCurrentTlId: function(tlid) {
             this.element.children("li").each(function() {
                 var $item = $(this),
-                    model = $item.data("model");
+                    model = $item.mopidyModel();
                 $item.find("a").toggleClass("ui-btn-active", model.tlid === tlid);
             });
             this._currentTlId = tlid;
